// $Id$
// Member functions for breakpoint distance graphs.
//
// Copyright (C) 1997 Humberto Ortiz Zuazaga
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// See the file licence.html for a copy of the GNU GPL

#include <time.h>

#include <LEDA/random.h>

#include "brkgraph.h"

#define SETLIM 50
#define CANDNUM 50
#define TRIALNUM 100

BrkGraph::BrkGraph(class map &mapdata, int minbrk) {

  // Allocate space for nodes and edges.
  n = mapdata.cols();
  nodes = new MNode[n];
  edges = new char *[n];

  // Make the nodes.
  for (int i = 0; i < n; i++) {
    nodes[i].setindex(i);
    nodes[i].setweight(mapdata.getmarker(i).getweight());
  }

  // Test and insert edges.
  int weight;
  for (int i = 0; i < n; i++) {
    edges[i] = new char[n];
    for (int j = 0; j < n; j++) {
      weight = mapdata.getpair(i,j);
      if (minbrk <= weight) {
	edges[i][j] = 1;
      } else {
	edges[i][j] = 0;
      }
    }
  }
}

BrkGraph::~BrkGraph(void) {
  delete [] nodes;
  delete [] edges;
}


int BrkGraph::rdfs(int *order, int cend, int tend, int best) {
  // Find the heavyest clique.
  // Partial enumeration algorithm cf. Applegate & Johnson
  // They derived it from Carraghan & Pardalos 1990.

//@misc{dfmax,
//         author = "David Applegate and David S. Johnson",
//         title = "dfmax, A Computer Program for the Maximum Clique
//         Problem",

//         howpublished = "Available via anonymous ftp from
//         {\texttt{<URL:ftp://dimacs.rutgers.edu/pub/challenge/graph/solvers/>}}",

//         note = "A benchmark program for the Second DIMACS Challenge on
//         Clique, Coloring and Satisfiability, Rutgers University,
//         October 11--13, 1993",
// }
 
  int weight = 0;
  int *neworder = new int[tend];

  for (int i = 0; i < cend; i++) {
    weight += nodes[order[i]].getweight();
    neworder[i] = order[i];
  }

  int testweight = 0;
  int mark = 0;

  for (int i = cend; i < tend; i++) {
    neworder[cend] = order[i];
    mark = cend + 1;

    testweight = weight + nodes[neworder[cend]].getweight();

    for (int j = i+1; j < tend; j++) {
      if (edges[neworder[cend]][order[j]]) {
	testweight += nodes[order[j]].getweight();
	neworder[mark++] = order[j];
      }
    }

    // if testweight viable, recurse
    if (best < testweight) {
      best = rdfs (neworder, cend+1, mark, best);
    }
  }

  // if we have new best, save.
  if (weight > best) {
    best = weight;
    cout << "Found clique of size " << cend << " , weight " << weight << endl;
    for (int i = 0; i < cend; i++) {
      cout << nodes[order[i]].getindex() + 1 << " ";
    }
    cout << endl;
  }

  delete [] neworder;

  return best;
}

// Implements Feo Resende & Smith alg for Max Ind Set.
// A Greedy Randomized Adaptive Search Procedure for Maximum Independent Set.
// (1994)
// Operations Research 42(5):860-878

int BrkGraph::FRS_MIS(int *order, int cend, int tend, int best) {
  cout << order[cend] << order[tend] << best << endl;
  return 0;
}

// Implement the maximum independent set subroutine from XRLF.
// Johnson, Aragon, McGeoch, Schevon (1991) Optimization by Simulated
// Annealing: an Experimental Evaluation; Part II, Graph Coloring and
// Number Partitioning.  Operations Research 39(3):378-406.

int BrkGraph::XRLF_MIS(int *oldorder, int oldcend, int oldtend, int best,
		       int TrialNum, int SetLim, int CandNum) {

  int bestdeg = -1;
  int bestvertex = -1;
  int bestweight = -1;
  int cand, cdeg, cweight;
  random_source Rand;
  Rand.set_seed(time(NULL));

  int cend = oldcend;
  int tend = oldtend;
  int *order = new int[tend];

  if (tend <= SetLim) {
    TrialNum = 1;
  }

  for (int t = 0; t < TrialNum; t++) {
    cend = oldcend;
    tend = oldtend;
    for (int i = 0; i < tend; i++) {
      order[i] = oldorder[i];
    }
    while (cend < tend) {
      if (SetLim >= (tend - cend)) {
	// Exhaustive search
	best = rdfs(order, cend, tend, best);
	// exit the while loop
	break;
      } else {
	bestdeg = -1; bestvertex = -1; bestweight = -1;
	for (int i = 0; i < CandNum; i++) {
	  // Pick a random test vertex
	  cand = Rand(cend, tend - 1);
	  cdeg = 0;
	  cweight = nodes[order[cand]].getweight();
	  for (int j = 0; j < tend; j++) {
	    // compute degree
	    if (edges[order[cand]][order[j]]) {
	      cdeg++;
	    }
	  } // end for j
	  // If degree better than best, update best.
	  if ((cweight > bestweight) && (cdeg >= bestdeg)) {
	    bestdeg = cdeg;
	    bestvertex = cand;
	    bestweight = cweight;
	  }
	} // end for i
    
	// Add best to clique
	int tmp = order[cend];
	order[cend] = order[bestvertex];
	order[bestvertex] = tmp;
	
	// Update test vertices
	int mark = cend+1;
	for (int k = mark; k < tend; k++) {
	  if (edges[order[cend]][order[k]]) {
	    order[mark++] = order[k];
	  }
	}
	cend++;
	tend = mark;
      } // endif setlim
    } // endwhile test markers
  } // end for t

  delete [] order;
  return best;
}
