// $Id: map.cc 1.3 Tue, 25 Feb 1997 12:34:46 -0400 humberto $
// Class member definitions for BPA mapping data.
//
// Copyright (C) 1995, 1996 Humberto Ortiz Zuazaga and Rosemarie Plaetke.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// See the file licence.html for a copy of the GNU GPL

#include <assert.h>
#include <ctype.h>
#include <iomanip.h>

#include "map.h"

// A helper function.
istream& eatwhite(istream &in) {
  char c;

  while (in.get(c)) {
    if (!isspace(c)) {
      in.putback(c);
      break;
    }
  }
  return in;
}

map::map(void) {
  markers = NULL;
  chromosomes = NULL;
  PMU = NULL;
  Pair = NULL;
}

int map::setpair(int i, int j, int score) {
  return (Pair[i][j] = score);
}

int map::pairscore (int i, int j) {
  int retval = 0;
  int index;

  if (i != j) {
    for (index = 0; index < rows(); index++) {
      if ((('M' == PMU[index][j]) && ('P' == PMU[index][i])) ||
	  (('M' == PMU[index][i]) && ('P' == PMU[index][j]))) {
	retval++;
      }
    }
  }
  Pair[i][j] = Pair[j][i] = retval;
  return retval;
}

istream& map::read (istream &in, int which) {
  in >> m;
  eatwhite(in);

  markers = new marker[m];
  for (int i = 0; i < m; i++) {
    string name;
    name.read(in, '\n');
    markers[i].setname(name);
  }

  in >> n;
  eatwhite(in);

  PMU = new string[n];
  for (int i = 0; i < n; i++) {
    PMU[i].read(in, '\n');
  }

  chromosomes = new chromosome[n];

  for (int i = 0; i < m; i++) {
    markers[i].setweight((KNOWNS == which) ? mrk_knowns(i) : mrk_retained(i));
  }

  for (int i = 0; i < n; i++) {
    chromosomes[i].setknowns(chr_knowns(i));
  }

  Pair = new int *[m];
  for (int i = 0; i < m; i++) {
    Pair[i] = new int[m];
    for (int j = 0; j <= i; j++) {
      pairscore(i, j);
    }
  }

  return in;
}

int map::mrk_knowns (int index) {
  int knowns = 0;

  for (int j = 0; j < n; j++) {
    if (PMU[j][index] != 'U') {
      knowns++;
    }
  }
  return knowns;
}

int map::mrk_retained (int index) {
  int knowns = 0;

  for (int j = 0; j < n; j++) {
    if (PMU[j][index] == 'P') {
      knowns++;
    }
  }
  return knowns;
}

int map::chr_knowns (int index) {
  int knowns = 0;

  for (int j = 0; j < m; j++) {
    if (PMU[index][j] != 'U') {
      knowns++;
    }
  }
  return knowns;
}

int map::RL_score(int *order) {
  int sum = 0;

  int i,j;
  char left;

  for (i = 0; i < n; i++) {
    left = 'U';
    for (j = 0; j < m; j++) {
      if (((left == 'P') && ('M' == PMU[i][order[j]])) ||
          ((left == 'M') && ('P' == PMU[i][order[j]]))) {
        sum++;
      }
      if (PMU[i][order[j]] != 'U') {
        left = PMU[i][order[j]];
      }
    }
  }
  return sum;
}

int map::NRL_score(int *order) {
  int sum = 0;

  int i;

  for (i = 0; i < m - 1; i++) {
    sum += Pair[order[i]][order[i+1]];
  }
  return sum;
}

int map::RL_score(int *order, int depth) {
  int sum = 0;

  int i,j;
  char left;

  for (i = 0; i < n; i++) {
    left = 'U';
    for (j = 0; j < depth; j++) {
      if (((left == 'P') && ('M' == PMU[i][order[j]])) ||
          ((left == 'M') && ('P' == PMU[i][order[j]]))) {
        sum++;
      }
      if (PMU[i][order[j]] != 'U') {
        left = PMU[i][order[j]];
      }
    }
  }
  return sum;
}

int map::NRL_score(int *order, int depth) {
  int sum = 0;

  int i;

  for (i = 0; i < depth - 1; i++) {
    sum += Pair[order[i]][order[i+1]];
  }
  return sum;
}

int map::bound(int *order, int depth, int len) {
  int sum = 0;
  int *neworder;

  int i,j,k;

  if (depth == 0) depth = 1;

  sum += NRL_score(order, depth);

  if (depth == len) return sum;

  // do prims
  neworder = new int[len];
  for (i = 0; i < len; i++) {
    neworder[i] = order[i];
  }

  for (i = depth; i < len; i++) {
    int min = Pair[neworder[i-1]][neworder[i]];
    int index = i;
    int cost;

    for (j = i; j < len; j++) {
      for (k = depth - 1; k < i; k++) {
	cost = Pair[neworder[k]][neworder[j]];
	if (min > cost) {
	  min = cost;
	  index = j;
	}
      }
    }
    int tmp = neworder[i];
    neworder[i] = neworder[index];
    neworder[index] = tmp;
    sum += min;
  }

  delete [] neworder;
  return sum;
}

marker& map::getmarker(int i) {
  return markers[i];
}

chromosome& map::getchromosome(int i) {
  return chromosomes[i];
}

void map::pretty(int *order) {
  for (int i = 0; i < m - 1; i++) {
    cout << markers[order[i]].getname() << " ";
  }
  cout << markers[order[m - 1]].getname() << endl;
}

void map::pretty(int *order, int len) {
  for (int i = 0; i < len - 1; i++) {
    cout << markers[order[i]].getname() << " ";
  }
  cout << markers[order[len - 1]].getname() << endl;
}
