// $Id: sa-local.cc,v 1.1 1996/08/21 16:05:30 humberto Exp humberto $
// Local search heuristics for BPA (based on TSP).

#include <limits.h>
#include <math.h>
#include <stdlib.h>
#include <fstream.h>
#include <time.h>
#include <stdio.h>

#include <LEDA/array.h>

#include "map.h"

int flip(int);
int *neighbor(int *, int);
int *start(int *, int);

int main (int argc, char *argv[]) {
  map mapdata;

  double T;

  if (argc < 3) {
    cerr << "Usage: sa-local {mapdata} {sequence}" << endl;
    exit(1);
  }

  ifstream in(argv[1]);
  if (!in) { 
    perror ("Error opening map file"); 
    exit (1);
  }
  mapdata.read(in, KNOWNS);
  in.close();

  // Read in the sequence file, this should be in a function or class!
  ifstream seqin(argv[2]);
  array<int> sequence(mapdata.cols());
  if (!seqin) {
    perror ("Error opening sequence file"); 
    exit (1);
  }
  sequence.read(seqin);
  seqin.close();

  int length;
  for (length = 0; length <= sequence.high(); length++) {
    // need way more error checking!
    if (sequence[length] == 0) {
      break;
    }
  }

  int *seq = new int[length];
  for (int i = 0; i < length; i++) {
    seq[i] = sequence[i] - 1;
  }

  int runlimit = 500 * length;
  int suclimit = 50 * length;
  if (4 == argc) {
    runlimit = atoi(argv[3]);
  }

  // Pick an initial order.
  int *order, *next;
  srandom(time(NULL));
  order = start(seq, length);
  int best = INT_MAX;
  int *save = new int[length];
  int myscore, nextscore;

  T = (double) mapdata.NRL_score(order,length) / (double) length;

  for (int k = 0; k < 1000; k++) {
    int success = 0;		// improved runs.
    for (int i = 0; i < runlimit; i++) {
      // save best order.
      myscore = mapdata.NRL_score(order, length);
      if (myscore < best) {
	best = myscore;
	for (int j = 0; j < length; j++) {
	  save[j] = order[j];
	}
      } else {
	// failed++;
      }
      
      next = neighbor(order, length);
      nextscore = mapdata.NRL_score(next, length);
      if (myscore >= nextscore) {
	success++;
	delete [] order;
	order = next;
      } else {
	double delta = myscore - nextscore;
	double P = exp(delta/T);
	double x = flip(RAND_MAX) * 1.0 / RAND_MAX;
	
	if (x <= P) {
	  success++;
	  delete [] order;
	  order = next;
	} else {
	  delete [] next;
	}      
      }

      if (success >= suclimit) {
	break;
      }
    } // end for runlimit

    printf ("best, T, success = %d %g %d\n", best, T, success);
    fflush(stdout);

    if (success == 0) {
      break;
    }
    T = T * 0.9;
    
  } // end for k
  
  
  // pretty print the map!
  cout << "Score = " << best << endl;
  for (int j = 0; j < length; j++) {
    cout << save[j] + 1 << " ";
  }
  cout << endl;
  // mapdata.pretty(save, length);
}

int flip(int sides) {
  long int r = random();
  return r % sides;
}

int *start(int *A, int len) {
  int *order = new int[len];

  for (int i = 0; i < len; i++) {
    order[i] = A[i];
  }
  return order;
}

int *neighbor(int *old, int len) {
  int *next = new int[len];
  int pos[3];
  int i;

  for (i = 0; i < 3; i++) {
    pos[i] = flip(len);
  }

  // sort the offsets
  int tmp;
  if (pos[0] > pos[1]) {
    tmp = pos[1];
    pos[1] = pos[0];
    pos[0] = tmp;
  }
  if (pos[1] > pos[2]) {
    tmp = pos[2];
    pos[2] = pos[1];
    pos[1] = tmp;
  }
  if (pos[0] > pos[1]) {
    tmp = pos[1];
    pos[1] = pos[0];
    pos[0] = tmp;
  }

  int mark = 0;
  for (i = 0; i < pos[0]; i++) {
    next[mark++] = old[i];
  }
  for (i = pos[1]; i <= pos[2]; i++) {
    next[mark++] = old[i];
  }

  if (flip(2)) {
    for (i = pos[0]; i < pos[1]; i++) {
      next[mark++] = old[i];
    }
  } else {
    for (i = pos[1] - 1; i >= pos[0]; i--) {
      next[mark++] = old[i];
    }
  }

  for (i = pos[2] + 1; i < len; i++) {
    next[mark++] = old[i];
  }

  return next;
}

